//---------------------------------------------------------------------------
#pragma once

#include <string>
#include <vector>
#include <map>
//---------------------------------------------------------------------------
typedef std::vector<std::string>   VECTOR_CSTR;
//---------------------------------------------------------------------------
class 
		#ifdef __BUILD_DirectXLS
			__declspec( dllexport )
		#else
			__declspec( dllimport )
		#endif
ExcelBinary
{
public:
	//---------------------------------------------------------------------------
	static int GetRowColRange( const char* pchOperand, char chColumnSeparator,
						int nIndexRowBase, int nIndexColBase,
						bool& fIsRange, int& nIndexRow1, int& nIndexCol1, int& nIndexRow2, int& nIndexCol2);
	static int GetRowCol( const char* pchAddress, int nIndexRowBase, int nIndexColBase, int& nIndexRow, int& nIndexCol);
	static void TrimFormulaLeft( const char*& pchString);
public:
    class CxNotifyHandler;

    class CxNotifyConsumer
	{
	public:
		virtual void Notify( CxNotifyHandler* caller) =0;
	};
    class CxNotifyHandler
	{
	public:
		CxNotifyHandler( CxNotifyConsumer*	pxNotifyConsumer)
		: m_pxNotifyConsumer(pxNotifyConsumer)
		{}
		void Notify()
		{
			if( m_pxNotifyConsumer)	m_pxNotifyConsumer->Notify(this);
		}
	protected:
		CxNotifyConsumer*	m_pxNotifyConsumer;
	};

public:
    //---------------------------------------------------------------------------
    class RANGE_ADDRESS;
    typedef std::vector<RANGE_ADDRESS>   VECTOR_RANGE_ADDRESS;
    typedef std::vector<WORD>   VECTOR_WORD;
    typedef std::vector<DWORD>  VECTOR_DWORD;
    class USTRING;
    typedef std::map< std::string, USTRING>  MAP_CSTR_STRING_8;
    typedef std::map< int, USTRING>   MAP_INT_STRING_8;
    class EXTERNNAME_F;
    typedef std::map< std::string, EXTERNNAME_F>  MAP_CSTR_EXTERNNAME_F;
    typedef std::map< int,  EXTERNNAME_F>  MAP_INT_EXTERNNAME_F;
    class NAME;
    typedef std::map< std::string, NAME>  MAP_CSTR_NAME;
    typedef std::map< int, NAME>   MAP_INT_NAME;
    class RGBA;
    typedef std::map< RGBA, int>   MAP_RGB_INT;
    typedef std::map< int, RGBA>   MAP_INT_RGB;
    enum EArgType { ARG_TYPE_REF, ARG_TYPE_VAL, ARG_TYPE_ARRAY};
    //---------------------------------------------------------------------------
    class BASE
    {
    public:
        WORD m_wID;
        int m_nArrayIndex;

    public:
        BASE( WORD wID);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount() =0;
        ExcelBinary::BASE& operator=( const ExcelBinary::BASE& base);

    public:
        int GetSize();
        bool IsModified() const;
        void SetNotifyOnChange( CxNotifyHandler* pxNotifyHandler) 
		{ m_pxNotifyHandler = pxNotifyHandler;}

    protected:
        bool m_fToSave;
        CxNotifyHandler* m_pxNotifyHandler;

    private:
        WORD m_wSize;

    };
    //---------------------------------------------------------------------------
    //---------------------------------------------------------------------------
    class BOF : BASE
    {
    public:
        WORD m_wVersion;
        WORD m_wType;
        WORD m_wBuildID;
        WORD m_wBuildYear;
        DWORD m_dwHistoryFlags;
        DWORD m_dwLowestVersion;

    public:
        BOF( WORD wType);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    class BOF_WB_Globals :  public BOF
    {
    public:
        BOF_WB_Globals() : BOF( 0x0005){}
    };
    //---------------------------------------------------------------------------
    class BOF_WSheet :  public BOF
    {
    public:
        BOF_WSheet() : BOF( 0x0010){}
    };
    //---------------------------------------------------------------------------
    class EOS :  public BASE
    {
    public:
        EOS() : BASE( 0x000A){}
        virtual WORD GetCount(){ return 0;}
    };
    //---------------------------------------------------------------------------
    class WINDOW1 : public BASE
    {
    public:
        WORD m_wPosHorz;
        WORD m_wPosVert;
        WORD m_wWidth;
        WORD m_wHeight;
        WORD m_wFlags;
        WORD m_wIndexWS;
        WORD m_wIndexTab;
        WORD m_wNumSelectedWS;
        WORD m_wWidthTabBar;

    public:
        WINDOW1();
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    enum ETypeUSTRING { USTRING_IS_8  = 0x00,
                        USTRING_IS_16 = 0x01,
                      };
    //---------------------------------------------------------------------------
    class USTRING
    {
    protected:
        WORD m_wLength;
        BYTE m_bFlags;
        char* m_pchString;
        WCHAR* m_pwchString;
        bool m_fIsLen16or8;
        bool m_fWOLength;
        std::string m_sString;

    public:
        int m_nIndex;// for internal use: in SST

    public:
        USTRING();
        USTRING( const char* lpszString, bool fIsChar8or16, bool fIsLen16or8, bool fWOLength =false);
        ~USTRING();

        bool Save( IStream*  pIStream) const;
        WORD GetCount();
        WORD GetLength();

        USTRING& operator=( const USTRING& ustr);
        bool operator==( const USTRING& ustr) const;
        bool operator!=( const USTRING& ustr) const	{return !operator==( ustr);}

    public:
        void SetString( const char* lpszString);

        const char*  GetString() const;
        const WCHAR* GetStringW() const;
    };
    //---------------------------------------------------------------------------
    class STRING_8
    {
    protected:
        WORD m_wLength;
        BYTE m_bFlags;
        std::string m_sString;

    public:
        int m_nIndex;// for internal use: in SST

    public:
        STRING_8( char* lpszString);
        STRING_8();
        ~STRING_8();

        bool Save( IStream*  pIStream);
        WORD GetCount();
    };
    //---------------------------------------------------------------------------
    enum FontStyles     {
                            FONT_STYLE_NONE = 0x0000,
                            FONT_STYLE_BOLD = 0x0001,
                            FONT_STYLE_ITALIC = 0x0002,
                            FONT_STYLE_UNDERLINED = 0x0004
                        };
    enum FontUnderLine  {
                            FONT_UNDERLINE_NONE = 0x00,
                            FONT_UNDERLINE_SINGLE = 0x01,
                            FONT_UNDERLINE_DOUBLE = 0x02
                        };
    //---------------------------------------------------------------------------
    class RGBA
    {
    protected:
        BYTE m_bRed;
        BYTE m_bGreen;
        BYTE m_bBlue;
        BYTE m_bNotUsed;

        bool m_fIsModified;

    public:
        RGBA();
        RGBA( BYTE bRed, BYTE bGreen, BYTE bBlue);
        RGBA( COLORREF refColor);
    public:
        RGBA( const RGBA& rgb){  *this = rgb;}

		RGBA& operator=( const RGBA& rgb);

        bool operator==( const RGBA& rgb) const;
        bool operator!=( const RGBA& rgb) const
        {
            return !(*this == rgb);
        }
        bool operator<( const RGBA& rgb) const;

    public:
        bool Save( IStream*  pIStream);
        WORD GetCount() const;
        COLORREF GetColor() const;

        bool IsModified();
    };
    //---------------------------------------------------------------------------

    //---------------------------------------------------------------------------
    class PALETTE : public BASE
    {
    protected:
        MAP_RGB_INT m_mapRGBInt;

    public:
        PALETTE() : BASE(0x0092){}

    public:
        bool Save( IStream*  pIStream);
        WORD GetCount();
    public:
        WORD SetColor( BYTE bRed, BYTE bGreen, BYTE bBlue);
        WORD SetColor( const RGBA& rgb);
    };
    //---------------------------------------------------------------------------
    class FONT : public BASE
    {
    protected:
        WORD m_wHeight;
        WORD m_wFlags;
        WORD m_wColorIndex;
        WORD m_wWeight;
        WORD m_wEscapement;
        BYTE m_bUndeline;
        BYTE m_bFamily;
        BYTE m_bCharacterSet;
        BYTE m_bReserved;
        USTRING m_ustrName;
        RGBA m_RGB;

    protected:
        void SetSize( WORD wSize, bool fInit);
        FontStyles GetStyles() const
        {
            return (FontStyles)m_wFlags;
        }
        void SetStyles( FontStyles eFontStyles, bool fInit);
        void SetName( const char* lpszName, bool fInit);
        void SetColor( BYTE bRed, BYTE bGreen, BYTE bBlue, bool fInit);
        void SetColorIndex( PALETTE& Palette, bool fInit);

    public:
        FONT( WORD wSize, FontStyles eFontStyles, const char* lpszName);
        FONT();
        FONT( const FONT& font) : BASE(0x0031){
                                                *this = font;
                                              }
        virtual ~FONT(){}

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

        bool operator==( const ExcelBinary::FONT& font) const;
        bool operator!=( const ExcelBinary::FONT& font) const
        {
            return !(*this == font);
        }
        //==============================================================================================
        void Assign( const ExcelBinary::FONT& font);
        void Inherit( const ExcelBinary::FONT& font);
        ExcelBinary::FONT& operator=( const ExcelBinary::FONT& font);
        //==============================================================================================

    public:
        void SetSize( WORD wSize);
        WORD GetSize() const;
        double GetHeight( bool fWithLineGap);

        void SetBold( bool fIsBold);
        bool IsBold() const;
        void SetUnderline( bool fIsUnderline);
        bool IsUnderline() const;
        void SetItalic( bool fIsItalic);
        bool IsItalic() const;
        void SetName( const char* lpszName);
        const char* GetName() const;

        void SetColor( COLORREF refColor);
        void SetColor( BYTE bRed, BYTE bGreen, BYTE bBlue);
        void SetColorIndex( PALETTE& Palette);
    };
    //---------------------------------------------------------------------------
    enum Align_Horz {
                        ALIGN_HORZ_GENERAL  = 0x00,
                        ALIGN_HORZ_LEFT     = 0x01,
                        ALIGN_HORZ_CENTER   = 0x02,
                        ALIGN_HORZ_RIGHT    = 0x03,
                        ALIGN_HORZ_FILLED   = 0x04,
                        ALIGN_HORZ_JUSTIFIED= 0x05
                    };
    enum Align_Vert {
                        ALIGN_VERT_TOP      = 0x00,
                        ALIGN_VERT_CENTER   = 0x01,
                        ALIGN_VERT_BOTTOM   = 0x02,
                        ALIGN_VERT_JUSTIFIED= 0x03
                    };
    enum Line_Styles{
                        LINE_STYLES_NONE    = 0x00,
                        LINE_STYLES_THIN    = 0x01,
                        LINE_STYLES_MEDIUM  = 0x02,
                        LINE_STYLES_THICK   = 0x05,
                        LINE_STYLES_DOUBLE  = 0x06,
                    };
    enum Assign_Flags{  ASSIGN_FLAGS_NORMAL          = 0,
                        ASSIGN_FLAGS_WITHOUT_LINES   = 1
                     };
    //---------------------------------------------------------------------------
    class XF : public BASE
    {
    public:
        WORD m_wIndexFont;
        WORD m_wIndexFormat;

    protected:
        WORD m_wIndexParent_Prot;
        BYTE m_bAlignVertWrapHorz;
    public:
        BYTE m_bRotateAngle;
    protected:
        BYTE m_bDirShrinkIndent;
        BYTE m_bUsedAttributes;
        DWORD m_dwBordersColorsA;
        DWORD m_dwBordersColorsB;
        WORD m_wPaternColors;
        RGBA m_RGB;

    protected:
        void SetFont( WORD wIndexFont, bool fToSetModified);
        void SetFormat( WORD wIndexFormat, bool fToSetModified);
        void SetAlignmentHorz( Align_Horz eAlign_Horz, bool fToSetModified);
        void SetAlignmentVert( Align_Vert eAlign_Vert, bool fToSetModified);
        void SetWrapText( bool fToWrap, bool fToSetModified);
        void SetLineLeft   ( Line_Styles eStyle, bool fToSetModified);
        void SetLineRight  ( Line_Styles eStyle, bool fToSetModified);
        void SetLineTop    ( Line_Styles eStyle, bool fToSetModified);
        void SetLineBottom ( Line_Styles eStyle, bool fToSetModified);
        void SetLineArround( Line_Styles eStyle, bool fToSetModified);
        void SetBackgroundColor( COLORREF refColor, bool fToSetModified);
        void SetBackgroundColor( BYTE bRed, BYTE bGreen, BYTE bBlue, bool fToSetModified);

    public:
        XF();
        XF( WORD wIndexFont, WORD wIndexFormat);
        XF( const XF& xf) : BASE(0x00E0){
                                                *this = xf;
                                              }
        virtual ~XF(){}
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
        //==============================================================================================
        void Assign( const ExcelBinary::XF& xf, Assign_Flags eFlags);
        void Inherit( const XF& xf);
        XF& XF::operator=( const XF& xf);
        //==============================================================================================

    protected:
        void Copy( const XF& xf, Assign_Flags eFlags);

    public:
        bool XF::operator==( const XF& xf) const;
        bool operator!=( const ExcelBinary::XF& xf) const	{ return !(*this == xf);}

    public:
        void SetParent( WORD wIndexParent);
        void MakeStyle();
            //---------------
        void SetDefault();
            //---------------
        void SetFont ( WORD wIndexFont){ SetFont( wIndexFont, true );}
        void CopyFont( WORD wIndexFont){ SetFont( wIndexFont, false);}
        WORD GetFont();
            //---------------
        void SetFormat ( WORD wIndexFormat){ SetFormat( wIndexFormat, true );}
        void CopyFormat( WORD wIndexFormat){ SetFormat( wIndexFormat, false);}
        WORD GetFormat();
            //---------------
        void SetAlignmentHorz( Align_Horz eAlign_Horz){ SetAlignmentHorz( eAlign_Horz, true );}
        void SetAlignmentVert( Align_Vert eAlign_Vert){ SetAlignmentVert( eAlign_Vert, true );}
        void SetWrapText( bool fToWrap){ SetWrapText( fToWrap, true );}
        void CopyAlignmentHorz( Align_Horz eAlign_Horz){ SetAlignmentHorz( eAlign_Horz, false);}
        void CopyAlignmentVert( Align_Vert eAlign_Vert){ SetAlignmentVert( eAlign_Vert, false);}
        void CopyWrapText( bool fToWrap){ SetWrapText( fToWrap, false);}

        Align_Horz GetAlignmentHorz() const;
        Align_Vert GetAlignmentVert() const;
        bool GetWrapText() const;
            //---------------
        void SetLineLeft   ( Line_Styles eStyle){ SetLineLeft   ( eStyle, true);}
        void SetLineRight  ( Line_Styles eStyle){ SetLineRight  ( eStyle, true);}
        void SetLineTop    ( Line_Styles eStyle){ SetLineTop    ( eStyle, true);}
        void SetLineBottom ( Line_Styles eStyle){ SetLineBottom ( eStyle, true);}
        void SetLineArround( Line_Styles eStyle){ SetLineArround( eStyle, true);}

        void CopyLineLeft   ( Line_Styles eStyle){ SetLineLeft   ( eStyle, false);}
        void CopyLineRight  ( Line_Styles eStyle){ SetLineRight  ( eStyle, false);}
        void CopyLineTop    ( Line_Styles eStyle){ SetLineTop    ( eStyle, false);}
        void CopyLineBottom ( Line_Styles eStyle){ SetLineBottom ( eStyle, false);}
        void CopyLineArround( Line_Styles eStyle){ SetLineArround( eStyle, false);}

        Line_Styles GetLineLeft() const  ;
        Line_Styles GetLineRight() const ;
        Line_Styles GetLineTop() const   ;
        Line_Styles GetLineBottom() const;
            //---------------
        void SetBackgroundColor( COLORREF refColor){ SetBackgroundColor( refColor, true);}
        void SetBackgroundColor( BYTE bRed, BYTE bGreen, BYTE bBlue){ SetBackgroundColor( bRed, bGreen, bBlue, true);}
        void SetBackgroundColorIndex( PALETTE& Palette);
        void CopyBackgroundColor( COLORREF refColor){ SetBackgroundColor( refColor, false);}
        void CopyBackgroundColor( BYTE bRed, BYTE bGreen, BYTE bBlue){ SetBackgroundColor( bRed, bGreen, bBlue, false);}
        void CopyBackgroundColorIndex( WORD wColorIndex);
        WORD GetBackgroundColorIndex();
        COLORREF GetBackgroundColor() const;
            //---------------
        bool IsFormatChanged()          ;
        bool IsFontChanged()            ;
        bool IsAlignmentWrapChanged()   ;
        bool IsLinesChanged()           ;
        bool IsBackgroundColorChanged() ;
    };
    //---------------------------------------------------------------------------
    class FORMAT : public BASE
    {
    protected:
        USTRING m_USTRING;

    public:
        FORMAT();

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

        FORMAT( const FORMAT& format) : BASE(0x041E){
                                                *this = format;
                                              }
        virtual ~FORMAT(){}
        void Assign( const ExcelBinary::FORMAT& format);
        void Inherit( const FORMAT& format);
        FORMAT& operator=( const FORMAT& format);
        bool FORMAT::operator==( const FORMAT& format) const;
        bool operator!=( const FORMAT& format) const	{ return !(*this == format);}

    public:
        void SetFormatString( const char* lpszString);
        const char*  GetFormatString();
        const WCHAR* GetFormatStringW();
    };
    //---------------------------------------------------------------------------
    class STYLE_BUILTIN : public BASE
    {
    public:
        WORD m_wIndexXF;
        BYTE m_bBuiltInID;
        BYTE m_bLevel;

    public:
        STYLE_BUILTIN( WORD wIndexXF, BYTE bBuiltInID);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    public:
        void SetIndexXF( WORD wIndexXF);
    };
    //---------------------------------------------------------------------------
    class COLINFO : public BASE
    {
    protected:
        WORD m_wIndexColFirst;
        WORD m_wIndexColLast;
        WORD m_wWidth;
        WORD m_wIndexXF;
        WORD m_wOptions;
        WORD m_wNotUsed;

    public:
        COLINFO();
        COLINFO( WORD wIndexColFirst, WORD wIndexColLast, WORD wIndexXF);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
        void SetRange( WORD wIndexColFirst, WORD wIndexColLast);
        void SetXF( WORD wIndexXF);
        void SetWidth( double dbWidthNumChars);
        void SetHidden( bool fHidden);
    };
    //---------------------------------------------------------------------------
    class RANGE_ADDRESS
    {
    public:
        WORD m_wRowFirst;
        WORD m_wRowLast;
        WORD m_wColFirst;
        WORD m_wColLast;

    public:
        RANGE_ADDRESS( WORD wRowFirst, WORD wRowLast, WORD wColFirst, WORD wColLast);
        RANGE_ADDRESS( const RANGE_ADDRESS& ra){  *this = ra;}

		RANGE_ADDRESS& operator=( const RANGE_ADDRESS& ra);

    public:
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    class MERGED_CELLS : public BASE
    {
    protected:
        VECTOR_RANGE_ADDRESS m_vecMerged;

    public:
        MERGED_CELLS() : BASE(0x00E5){}

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
        void SetMergedRange( WORD wRowFirst, WORD wColFirst, WORD wRowLast, WORD wColLast);
        bool IsEmpty(){ return m_vecMerged.size() == 0;}
    };
    //---------------------------------------------------------------------------
    class BOUNDSHEET : public BASE
    {
    public:
        DWORD m_dwSheetBOFOffset;
        BYTE m_bVisibility;
        BYTE m_bSheetType;
        USTRING m_ustrSheetName;

    public:
        BOUNDSHEET( DWORD dwSheetBOFOffset, const char* lpszSheetName);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

        void SetOffset( DWORD dwSheetBOFOffset);
    };
    //---------------------------------------------------------------------------
    class DIMENSIONS : public BASE
    {
    protected:
        DWORD m_dwIndexFirstRow;
        DWORD m_dwIndexLastRow;
        WORD m_wIndexFirstCol;
        WORD m_wIndexLastCol;
        WORD m_wReserved;

    public:
        DIMENSIONS( DWORD dwIndexFirstRow, DWORD dwIndexLastRow,
                        WORD wIndexFirstCol, WORD wIndexLastCol);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    class WINDOW2 : public BASE
    {
    protected:
        WORD m_wFlags;
        WORD m_wIndexFirstVisibleRow;
        WORD m_wIndexFirstVisibleCol;
        WORD m_wGridColorIndex;
        WORD m_wReserved;
        WORD m_wFactorPreview;
        WORD m_wFactorNormalView;
        DWORD m_dwReserved;

    public:
        WINDOW2( bool fIsSheetSelected);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    class INDEX : public BASE
    {
    protected:
        DWORD m_dwReserved;
        DWORD m_dwFirstRowUsed;
        DWORD m_dwFirstRowUnused;
        DWORD m_dwOffsetDefColWidth;
        VECTOR_DWORD m_vecOffsetDBCells;

    public:
        INDEX(  DWORD dwFirstRowUsed, DWORD dwFirstRowUnused,
                      DWORD dwOffsetDefColWidth, VECTOR_DWORD vecOffsetDBCells);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    class ROW : public BASE
    {
    protected:
        WORD m_wIndexRow;
        WORD m_wIndexColFirst;
        WORD m_wIndexColLast;
        WORD m_wHeight;
        WORD m_wReserved;
        WORD m_wReservedB;
        DWORD m_dwFlags;

    public:
        ROW( WORD wIndexRow, WORD wIndexColFirst, WORD wIndexColLast, int nIndexXF);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

        void SetHeight( double dbPixels);
        void SetHidden( bool fHidden);
        void SetFormat( WORD wIndexXF);
    };
    //---------------------------------------------------------------------------
    //---------------------------------------------------------------------------
    class FUNCTION
    {
    public:
        const std::string m_sName;
        const WORD m_wIndex;
        const bool m_fIsFixed;
        const bool m_fIsVar;
        const bool m_fIsExternal;
        const int m_nArgTypeAsThis;

    public:
        FUNCTION();
        FUNCTION( const std::string& sName, WORD wIndex, bool fIsFixed, bool fIsVar, bool fIsExternal, int nArgTypeAsThis);
        FUNCTION( const FUNCTION& aFunc);
        FUNCTION& operator=( const FUNCTION& aFunc);

        void AppendArgType( EArgType eType){    m_vecArgTypes.push_back( eType);}
        EArgType GetArgType( int nArgIndex) const;

    protected:
        std::vector<EArgType> m_vecArgTypes;
    };
    //---------------------------------------------------------------------------
    class FUNC_ARRAY : public std::map< std::string, FUNCTION>
    {
    };
    //---------------------------------------------------------------------------
    class DATAITEM : public BASE
    {
    public:
        WORD m_wIndexRow;
        WORD m_wIndexCol;
        WORD m_wIndexXF;

    public:
        DATAITEM( WORD wID, WORD wIndexRow, WORD wIndexCol, WORD wIndexXF);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    public:
        void SetAddress( WORD wIndexRow, WORD wIndexCol);
        void SetXF( WORD wIndexXF);
    };
    //---------------------------------------------------------------------------
    class BLANK : public DATAITEM
    {
    public:
        BLANK( WORD wIndexRow, WORD wIndexCol, WORD wIndexXF)
        : DATAITEM(0x0201, wIndexRow, wIndexCol, wIndexXF){}
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount(){ return DATAITEM::GetCount();}
    };
    //---------------------------------------------------------------------------
    class LABEL : public DATAITEM
    {
    public:
        USTRING m_uStr;

    public:
        LABEL( char* lpszStr, WORD wIndexRow, WORD wIndexCol, WORD wIndexXF)
        : DATAITEM(0x0204, wIndexRow, wIndexCol, wIndexXF), m_uStr( lpszStr, false, false){}
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount(){ return DATAITEM::GetCount() + m_uStr.GetCount();}
    };
    //---------------------------------------------------------------------------
    class CONTINUE : public BASE
    {
    protected:
        WORD m_wSize;

    public:
        CONTINUE( WORD wSize) : BASE( 0x003C){  m_wSize = wSize;}

        virtual bool Save( IStream*  pIStream);

        virtual WORD GetCount(){ return m_wSize;}
    };
    //---------------------------------------------------------------------------
    class SST : public BASE
    {
    public:
        SST();

        virtual bool Save( IStream*  pIStream);
        int SetString( const char* lpszString);

    protected:
        virtual WORD GetCount();
        virtual WORD GetCount( MAP_INT_STRING_8::iterator p, MAP_INT_STRING_8::iterator& pNext);

    protected:
        DWORD m_dwCountTotal;
        MAP_CSTR_STRING_8 m_mapStrings;
        MAP_INT_STRING_8 m_mapIntString16;
    };
    //---------------------------------------------------------------------------
    class LABELSST : public DATAITEM
    {
    public:
        DWORD m_dwIndexSST;

    public:
        LABELSST();
        LABELSST( DWORD dwIndexSST, WORD wIndexRow, WORD wIndexCol, WORD wIndexXF);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    public:
        void SetSST( DWORD dwIndexSST);
    };
    //---------------------------------------------------------------------------
    class NUMBER : public DATAITEM
    {
    public:
        double m_dbNumber;

    public:
        NUMBER();
        NUMBER( double dbNumber, WORD wIndexRow, WORD wIndexCol, WORD wIndexXF);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
        void SetNumber( double dbNumber);
    };
    //---------------------------------------------------------------------------
    class SUPBOOK;
    //---------------------------------------------------------------------------
    class FORMULA : public DATAITEM
    {
    protected:
        class TOKEN;
        class tArray;
    public:
		class CVTokensPtr : public std::vector< TOKEN*>
		{
		public:
			bool SaveTokens( IStream*  pIStream) const;
			WORD GetTokensCount( bool fToAddTwoBytesToSaveThisCount, bool fWithoutAdditional) const;
		};

    public:
        FORMULA();
        FORMULA( WORD wIndexRow, WORD wIndexCol, WORD wIndexXF, char chListSeparator, char chColumnSeparator);
        ~FORMULA();

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    public:
        void SetSeparators( char chListSeparator, char chColumnSeparator, char chDecimalSeparator);
        bool SetFormula( const char* pchFormula, int nIndexRowBase, int nIndexColBase,
                         const VECTOR_CSTR& vecSheets, SUPBOOK& SUPBOOK, const FUNC_ARRAY& aFuncArray,
                         bool fIsArray, WORD wRangeRowFirst, WORD wRangeColFirst, WORD wRangeRowLast, WORD wRangeColLast);

    protected:
        CVTokensPtr m_vecTokensPtr;
        tArray* m_ptArray;
        EArgType m_eArgType;
        const char* m_pchBegining;
        const VECTOR_CSTR* m_pvecSheets;
        SUPBOOK* m_pSUPBOOK;
        const FUNC_ARRAY* m_pFuncArray;

        bool m_fIsFormulaArray;
        WORD m_wRangeRowFirst;
        WORD m_wRangeColFirst;
        WORD m_wRangeRowLast;
        WORD m_wRangeColLast;

        char m_chListSeparator;
        char m_chColumnSeparator;
        char m_chDecimalSeparator;
        int m_nIndexRowBase;
        int m_nIndexColBase;

    protected:
        int SetFormulaCore( const char*& pchFormula, bool fGetFirstOpd);
        int AddOperand( const char*& pchFormula);
        int AddAllOperandsFunc( const char*& pchFuncArg, BYTE& bNumArg, const ExcelBinary::FUNCTION* pFunc);

        const ExcelBinary::FUNCTION& GetFunction( const char* pchFuncName);

        int AddAddress( const char* pchOperand, int nIndexSheet);
        bool IsUnary( const char* pchString);
        bool IsOperation( const char* pchFormula);
        int GetPriority( const char* pchFormula);
        TOKEN* GetOperation( const char*& pchFormula);
        const char* FindOperation( const char* pchFormula);
        bool AddToken( TOKEN* pToken);
        void ClearTokens();
        bool IsArray(){ return m_ptArray != NULL;}
        bool IsEnd_Full( const char* pchFormula);
        bool IsEnd_Paren( const char* pchFormula);
        bool IsFirstInArray(){  return m_fIsFormulaArray  &&
                                       (m_wRangeRowFirst == m_nIndexRowBase  &&  m_nIndexColBase ==  m_wRangeColFirst);}
        bool IsSecondInArray(){ return m_fIsFormulaArray  &&
                                       (m_wRangeRowFirst != m_nIndexRowBase  ||  m_nIndexColBase !=  m_wRangeColFirst);}

    protected:
            //---------------------------------------------------------------------------
            class TOKEN
            {
            protected:
                BYTE m_bID;

            public:
                TOKEN( BYTE bID){ m_bID = bID;}

                inline 
				virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return 1;}

                BYTE GetID(){ return m_bID;}
            };
            //---------------------------------------------------------------------------
            class tMinus : public TOKEN
            {
            public:
                tMinus() : TOKEN(0x13){}
            };
            //---------------------------------------------------------------------------
            class tPercent : public TOKEN
            {
            public:
                tPercent() : TOKEN(0x14){}
            };
            //---------------------------------------------------------------------------
            class tParen : public TOKEN
            {
            public:
                tParen() : TOKEN(0x15){}
            };
            //---------------------------------------------------------------------------
            class tNum : public TOKEN
            {
            protected:
                double m_dbNum;

            public:
                tNum( double dbNum) : TOKEN(0x1F), m_dbNum( dbNum){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 8;}
            };
            //---------------------------------------------------------------------------
            class tBool : public TOKEN
            {
            protected:
                bool m_fBool;

            public:
                tBool( bool fBool) : TOKEN(0x1D), m_fBool( fBool){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 1;}
            };
            //---------------------------------------------------------------------------
            class tStr : public TOKEN
            {
            protected:
                USTRING m_usStr;

            public:
                tStr( char* lpszStr) : TOKEN(0x17), m_usStr( lpszStr, false, false)  {}

                virtual bool Save( IStream*  pIStream) const {
                                                        TOKEN::Save( pIStream);
                                                        m_usStr.Save( pIStream);
                                                        return true;
                                                     }

                virtual WORD GetCount(){ return TOKEN::GetCount() + m_usStr.GetCount();}
            };
            //---------------------------------------------------------------------------
            class tRef : public TOKEN
            {
            protected:
                WORD m_wIndexRow;
                WORD m_wIndexCol;

            public:
                tRef( BYTE bID, WORD wIndexRow, WORD wIndexCol)
                : TOKEN( bID)                       { m_wIndexRow = wIndexRow;
                                                            m_wIndexCol = wIndexCol | 0xC000;}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 4;}
            };
            //---------------------------------------------------------------------------
            class tRefR : public tRef
            {
            public:
                tRefR( WORD wIndexRow, WORD wIndexCol) : tRef( 0x24, wIndexRow, wIndexCol){}
            };
            //---------------------------------------------------------------------------
            class tRefV : public tRef
            {
            public:
                tRefV( WORD wIndexRow, WORD wIndexCol) : tRef( 0x44, wIndexRow, wIndexCol){}
            };
            //---------------------------------------------------------------------------
            class tRefA : public tRef
            {
            public:
                tRefA( WORD wIndexRow, WORD wIndexCol) : tRef( 0x64, wIndexRow, wIndexCol){}
            };
            //---------------------------------------------------------------------------
            class tRef3d : public tRef
            {
            protected:
                WORD m_wIndexREF;

            public:
                tRef3d( BYTE bID, WORD wIndexREF, WORD wIndexRow, WORD wIndexCol)
                : tRef( bID, wIndexRow, wIndexCol){
                                                            m_wIndexREF = wIndexREF;
                                                        }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return tRef::GetCount() + 2;}
            };
            //---------------------------------------------------------------------------
            class tRef3dR : public tRef3d
            {
            public:
                tRef3dR( WORD wIndexREF, WORD wIndexRow, WORD wIndexCol)
                : tRef3d( 0x3A, wIndexREF, wIndexRow, wIndexCol){}
            };
            //---------------------------------------------------------------------------
            class tRef3dV : public tRef3d
            {
            public:
                tRef3dV( WORD wIndexREF, WORD wIndexRow, WORD wIndexCol)
                : tRef3d( 0x5A, wIndexREF, wIndexRow, wIndexCol){}
            };
            //---------------------------------------------------------------------------
            class tRef3dA : public tRef3d
            {
            public:
                tRef3dA( WORD wIndexREF, WORD wIndexRow, WORD wIndexCol)
                : tRef3d( 0x7A, wIndexREF, wIndexRow, wIndexCol){}
            };
            //---------------------------------------------------------------------------
            class tArea : public TOKEN
            {
            protected:
                WORD m_wIndexRow1;
                WORD m_wIndexCol1;
                WORD m_wIndexRow2;
                WORD m_wIndexCol2;

            public:
                tArea( BYTE bID, WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : TOKEN( bID)                       {
                                                        m_wIndexRow1 = wIndexRow1;
                                                        m_wIndexCol1 = wIndexCol1 | 0xC000;
                                                        m_wIndexRow2 = wIndexRow2;
                                                        m_wIndexCol2 = wIndexCol2 | 0xC000;
                                                    }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 8;}
            };
            //---------------------------------------------------------------------------
            class tAreaR : public tArea
            {
            public:
                tAreaR( WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea( 0x25, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2){}
            };
            //---------------------------------------------------------------------------
            class tAreaV : public tArea
            {
            public:
                tAreaV( WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea( 0x45, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2){}
            };
            //---------------------------------------------------------------------------
            class tAreaA : public tArea
            {
            public:
                tAreaA( WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea( 0x65, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2){}
            };
            //---------------------------------------------------------------------------
            class tArea3d : public tArea
            {
            protected:
                WORD m_wIndexREF;

            public:
                tArea3d( BYTE bID, WORD wIndexREF, WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea( bID, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2)
                                                    {
                                                        m_wIndexREF = wIndexREF;
                                                    }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 10;}
            };
            //---------------------------------------------------------------------------
            class tArea3dR : public tArea3d
            {
            public:
                tArea3dR( WORD wIndexREF, WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea3d( 0x3B, wIndexREF, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2){}
            };
            //---------------------------------------------------------------------------
            class tArea3dV : public tArea3d
            {
            public:
                tArea3dV( WORD wIndexREF, WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea3d( 0x5B, wIndexREF, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2){}
            };
            //---------------------------------------------------------------------------
            class tArea3dA : public tArea3d
            {
            public:
                tArea3dA( WORD wIndexREF, WORD wIndexRow1, WORD wIndexCol1, WORD wIndexRow2, WORD wIndexCol2)
                : tArea3d( 0x7B, wIndexREF, wIndexRow1, wIndexCol1, wIndexRow2, wIndexCol2){}
            };
            //---------------------------------------------------------------------------
            class exCachedValue : public TOKEN
            {
            protected:
                exCachedValue( BYTE bID) : TOKEN( bID){}
            };
            //---------------------------------------------------------------------------
            class exCachedNum : public exCachedValue
            {
            protected:
                double m_dbNum;

            public:
                exCachedNum( double dbNum) : exCachedValue( 0x01), m_dbNum( dbNum){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return exCachedValue::GetCount() + 8;}
            };
            //---------------------------------------------------------------------------
            class exCachedStr : public exCachedValue
            {
            protected:
                USTRING m_usStr;

            public:
                exCachedStr( char* lpszStr) : exCachedValue(0x02), m_usStr( lpszStr, true, true)  {}

                virtual bool Save( IStream*  pIStream) const {
                                                        exCachedValue::Save( pIStream);
                                                        m_usStr.Save( pIStream);
                                                        return true;
                                                     }

                virtual WORD GetCount(){ return exCachedValue::GetCount() + m_usStr.GetCount();}
            };
            //---------------------------------------------------------------------------
            class exCachedBool : public exCachedValue
            {
            protected:
                bool m_fBool;

            public:
                exCachedBool( bool fBool) : exCachedValue(0x04), m_fBool( fBool){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return exCachedValue::GetCount() + 8;}
            };
            //---------------------------------------------------------------------------
            class exCachedError : public exCachedValue
            {
            protected:
                BYTE m_bErrorCode;

            public:
                exCachedError( BYTE bErrorCode) : exCachedValue(0x10), m_bErrorCode( bErrorCode){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return exCachedValue::GetCount() + 8;}
            };
            //---------------------------------------------------------------------------
            typedef std::vector<exCachedValue*> ARRAY_ROW;
            typedef std::vector<ARRAY_ROW*>     ARRAY_MATRIX;
            //---------------------------------------------------------------------------
            class tArray : public TOKEN
            {
            public:
                tArray( BYTE bID) : TOKEN( bID){}
                ~tArray();

                virtual bool Save( IStream*  pIStream) const;
                virtual bool SaveMatrix( IStream*  pIStream) const;

                virtual WORD GetCount();
                WORD GetCountWithoutMatrix();

            public:
                void AddRow();
                void AddValue( exCachedValue* pValue);

            protected:
                ARRAY_MATRIX m_Matrix;

            protected:
                bool GetNumColsRows( BYTE& bNumCols, WORD& wNumRows) const;
                virtual WORD GetCountMatrix();
            };
            //---------------------------------------------------------------------------
            class tArrayR : public tArray
            {
            public:
                enum EID { ID = 0x20};
                tArrayR() : tArray( ID){}
            };
            //---------------------------------------------------------------------------
            class tArrayV : public tArray
            {
            public:
                enum EID { ID = 0x40};
                tArrayV() : tArray( ID){}
            };
            //---------------------------------------------------------------------------
            class tArrayA : public tArray
            {
            public:
                enum EID { ID = 0x60};
                tArrayA() : tArray( ID){}
            };
            //---------------------------------------------------------------------------
            class tFunc : public TOKEN
            {
            protected:
                WORD m_wIndexFunc;

            public:
                tFunc( BYTE bID, WORD wIndexFunc) : TOKEN( bID), m_wIndexFunc( wIndexFunc){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 2;}
            };
            //---------------------------------------------------------------------------
            class tFuncFixedV : public tFunc
            {
            public:
                tFuncFixedV( WORD wIndexFunc) : tFunc(0x41, wIndexFunc){}
            };
            //---------------------------------------------------------------------------
            class tFuncVarV : public tFunc
            {
            protected:
                BYTE m_bNumArg;

            public:
                tFuncVarV( WORD wIndexFunc, BYTE bNumArg)
                : tFunc( 0x42, wIndexFunc){
                                                  m_bNumArg = bNumArg;
                                              }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return 1 + 1 + 2;}
            };
            //---------------------------------------------------------------------------
            class tFuncOneV : public tFunc
            {
            protected:
                BYTE m_bFlags;

            public:
                tFuncOneV( WORD wIndexFunc, BYTE bNumArg)
                : tFunc( 0x19, wIndexFunc){
                                              m_bFlags = 0x10;
                                          }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return 1 + 1 + 2;}
            };
            //---------------------------------------------------------------------------
            class tName : public TOKEN
            {
            protected:
                WORD m_wIndexName;
                WORD m_wNotUsed;

            public:
                tName( BYTE bID, WORD wIndexName) : TOKEN( bID){
                                                        m_wIndexName = wIndexName;
                                                        m_wNotUsed = 0;
                                                     }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 2 + 2;}
            };
            //---------------------------------------------------------------------------
            class tNameR : public tName
            {
            public:
                tNameR( WORD wIndexName) : tName(0x23, wIndexName){}
            };
            //---------------------------------------------------------------------------
            class tNameX : public TOKEN
            {
            protected:
                WORD m_wIndexREF;
                WORD m_wIndexExtName;
                WORD m_wNotUsed;

            public:
                tNameX( BYTE bID, WORD wIndexREF, WORD wIndexExtName) : TOKEN( bID){
                                                        m_wIndexREF = wIndexREF;
                                                        m_wIndexExtName = wIndexExtName;
                                                        m_wNotUsed = 0;
                                                     }

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 2 + 2 + 2;}
            };
            //---------------------------------------------------------------------------
            class tNameXR : public tNameX
            {
            public:
                tNameXR( WORD wIndexREF, WORD wIndexExtName) : tNameX(0x39, wIndexREF, wIndexExtName){}
            };
            //---------------------------------------------------------------------------
            class tMacro : public tFuncVarV
            {
            protected:
                int m_nIndexName;

            public:
                tMacro( const char* lpszMacroName, BYTE bNumArg, SUPBOOK& supBook)
                : tFuncVarV( 0x00FF, 1 + bNumArg){}

                virtual bool Save( IStream*  pIStream) const {
                                                        tFuncVarV::Save( pIStream);
                                                        return true;
                                                     }

                virtual WORD GetCount(){ return tFuncVarV::GetCount();}
            };
            //---------------------------------------------------------------------------
            class tExp : public TOKEN
            {
            public:
                tExp( WORD wIndexRowFirst, WORD wIndexColFirst)
                : TOKEN( 0x01), m_wIndexRowFirst( wIndexRowFirst), m_wIndexColFirst( wIndexColFirst){}

                virtual bool Save( IStream*  pIStream) const;
                virtual WORD GetCount(){ return TOKEN::GetCount() + 2 + 2;}

            protected:
                WORD m_wIndexRowFirst;
                WORD m_wIndexColFirst;
            };
            //---------------------------------------------------------------------------
            //---------------------------------------------------------------------------
            class ARRAY : public BASE
            {
            public:
                ARRAY( WORD wIndexRowFirst, BYTE bIndexColFirst,
                         WORD wIndexRowLast, BYTE bIndexColLast, const CVTokensPtr& vecTokensPtr);
                virtual bool Save( IStream*  pIStream);
                virtual WORD GetCount();

            protected:
                WORD m_wIndexRowFirst;
                BYTE m_bIndexColFirst;
                WORD m_wIndexRowLast;
                BYTE m_bIndexColLast;
                const CVTokensPtr& m_vecTokensPtr;
            };
            //---------------------------------------------------------------------------
    };
    //---------------------------------------------------------------------------
    class DBCELL : public BASE
    {
    protected:
        DWORD m_dwOffsetFirstRow;
        VECTOR_WORD m_vecOffsetFirstCols;

    public:
        DBCELL( DWORD dwOffsetFirstRow, VECTOR_WORD& vecOffsetFirstCols);
        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
    };
    //---------------------------------------------------------------------------
    class REF
    {
    public:
        REF( WORD wIndexWS, WORD wIndexSUPBOOK);

        bool Save( IStream*  pIStream);
        WORD GetCount();

    public:
        WORD GetIndexWS();
        WORD GetIndexSUPBOOK();

    protected:
        WORD m_wIndexSUPBOOK;
        WORD m_wIndexWSFirst;
        WORD m_wIndexWSLast;
    };

    typedef std::vector<REF> CVSheetREF;
    //---------------------------------------------------------------------------
    class EXTERNSHEET : public BASE
    {
    protected:
        CVSheetREF m_vecSheetREF;

    public:
        EXTERNSHEET() : BASE(0x0017){}

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();
        bool IsEmpty(){ return (m_vecSheetREF.size() == 0);}

        WORD GetREF( WORD wIndexWS, WORD wIndexSUPBOOK =0);
        WORD GetExternalREF();
    };
    //---------------------------------------------------------------------------
    class EXTERNNAME_F : public BASE
    {
    public:
        EXTERNNAME_F( const char* lpszName);
        EXTERNNAME_F();
        EXTERNNAME_F& operator=( const EXTERNNAME_F& name);

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    protected:
        WORD  m_wFlags;
        DWORD m_dwNotUsed;
        USTRING m_usName;
        WORD m_wTrail;
    };
    //---------------------------------------------------------------------------
    class NAME : public BASE
    {
    public:
        NAME( const char* lpszName, WORD wIndexWS_Zero, WORD wFlags);
        NAME();
        NAME& operator=( const NAME& name);

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    protected:
        WORD m_wFlags;
        BYTE m_bShortCut;
        WORD m_wNotUsed;
        WORD m_wIndexWS_Zero; // zero for a global name
        USTRING m_usName;
    };
    //---------------------------------------------------------------------------
    class SUPBOOK : public BASE
    {
    protected:
        WORD m_wNumOfSheets;
        MAP_CSTR_EXTERNNAME_F m_mapExtNames_F;
        EXTERNSHEET m_ExternSheet;
        MAP_CSTR_NAME m_mapNames;

    public:
        SUPBOOK();

        virtual bool Save( IStream*  pIStream);
        virtual WORD GetCount();

    public:
        void SetNumOfSheets( WORD wNumOfSheets);
        WORD GetREF( WORD wIndexWS);
        WORD GetExternalREF();
        WORD SetExternalName( const char* lpszString);
        WORD SetName( const char* lpszString);
    };
    //---------------------------------------------------------------------------

};
//---------------------------------------------------------------------------
